@@ -110,9 +110,9 @@ The names of all the available application state variables can be determined wit
 :meth:`~MainAppBase.app_state_keys`.
 
 If your application is e.g. supporting a user-defined font size, using the provided/pre-defined app state variable
-:attr:`~MainAppBase.font_size`, then it has to call the method :meth:`change_app_state` with the
-:paramref:`~MainAppBase.change_app_state.state_name` set to `font_size` every time when the user has changed the font
-size of your app.
+:attr:`~MainAppBase.font_size`, then it has to call the method :meth:`change_app_state` with the argument of
+:paramref:`~MainAppBase.change_app_state.app_state_name` set to `font_size` every time when the user has changed the
+font size of your app.
 
 .. hint::
     The two built-in app state variables are :attr:`~MainAppBase.flow_id` and :attr:`~MainAppBase.flow_path` will be
@@ -156,8 +156,8 @@ app state events
 
 There are three types of notification events get fired in relation to the app state variables, using the method names:
 
-* `on_<state_name>`: fired if the user of the app is changing the value of an app state variable.
-* `on_<state_name>_save`: fired if a app state gets saved to the config file.
+* `on_<app_state_name>`: fired if the user of the app is changing the value of an app state variable.
+* `on_<app_state_name>_save`: fired if a app state gets saved to the config file.
 * `on_app_state_version_upgrade`: fired if the user upgrades an previously installed app to a higher version.
 
 The method name of the app state change notification event consists of the prefix ``on_`` followed by the variable name
@@ -378,18 +378,21 @@ from copy import deepcopy
 from typing import Any, Callable, Dict, List, Optional, Tuple, Type
 
 from ae.base import (                                                                                   # type: ignore
-    CFG_EXT, INI_EXT, NAME_PARTS_SEP, UNSET, duplicates, instantiate_config_parser, now_str, snake_to_camel)
+    CFG_EXT, INI_EXT, NAME_PARTS_SEP, UNSET,
+    duplicates, instantiate_config_parser, now_str, os_platform, snake_to_camel)
 from ae.files import RegisteredFile                                                                     # type: ignore
 from ae.paths import (                                                                                  # type: ignore
-    add_common_storage_paths, copy_file, copy_tree, norm_path, path_folders, Collector, FilesRegister, placeholder_key)
+    add_common_storage_paths, copy_file, copy_tree, norm_path, path_folders, path_name, placeholder_path,
+    Collector, FilesRegister, placeholder_key)
 from ae.updater import MOVES_SRC_FOLDER_NAME, check_all                                                 # type: ignore
 from ae.inspector import stack_var                                                                      # type: ignore
-from ae.i18n import default_language, load_language_texts, register_translations_path                   # type: ignore
+from ae.i18n import (                                                                                   # type: ignore
+    default_language, get_f_string, get_text, load_language_texts, register_translations_path)
 from ae.core import DEBUG_LEVELS, registered_app_names                                                  # type: ignore
 from ae.console import ConsoleApp                                                                       # type: ignore
 
 
-__version__ = '0.1.65'
+__version__ = '0.1.69'
 
 
 AppStateType = Dict[str, Any]                   #: app state config variable type
@@ -422,6 +425,11 @@ FLOW_KEY_SEP = ':'                              #: separator character between f
 FLOW_ACTION_RE = re.compile("[a-z0-9]+")        #: regular expression detecting invalid characters in flow action string
 FLOW_OBJECT_RE = re.compile("[A-Za-z0-9_]+")    #: regular expression detecting invalid characters in flow object string
 
+HIDDEN_GLOBALS = (
+    'ABC', 'abstractmethod', '_add_base_globals', 'Any', '__builtins__', '__cached__', 'Callable', '_d_', 'Dict',
+    '__doc__', '__file__', 'List', '__loader__', 'module_globals', '__name__', 'Optional', '__package__', '__path__',
+    '__spec__', 'Tuple', 'Type', '__version__')
+""" tuple of global/module variable names that are hidden in :meth:`~MainAppBase.global_variables` """
 
 PORTIONS_IMAGES = FilesRegister()               #: register of image files found in portions/packages at import time
 PORTIONS_SOUNDS = FilesRegister()               #: register of sound files found in portions/packages at import time
@@ -490,6 +498,21 @@ def flow_change_confirmation_event_name(flow_id: str) -> str:
     return f'on_{obj}_{action}'
 
 
+def flow_class_name(flow_id: str, name_suffix: str) -> str:
+    """ determine class name for the given flow id and class name suffix.
+
+    :param flow_id:         flow id.
+    :param name_suffix:     class name suffix.
+    :return:                name of the class. Please note that the flow action `open` will not be added
+                            to the returned class name.
+    """
+    flow, _index = flow_key_split(flow_id)
+    action, obj = flow_action_split(flow)
+    if action == 'open':
+        action = ''
+    return f'{snake_to_camel(obj)}{action.capitalize()}{name_suffix}'
+
+
 def flow_key(flow_id: str) -> str:
     """ return the key of a flow id.
 
@@ -521,6 +544,30 @@ def flow_object(flow_id: str) -> str:
     return flow_action_split(flow_key_split(flow_id)[0])[1]
 
 
+def flow_path_id(flow_path: List[str], path_index: int = -1) -> str:
+    """ determine the flow id of the last/newest entry in the flow_path.
+
+    :param flow_path:           flow path to get the flow id from.
+    :param path_index:          index in the flow_path.
+    :return:                    flow id string or empty string if flow path is empty or index does not exist.
+    """
+    if len(flow_path) >= (abs(path_index) if path_index < 0 else path_index + 1):
+        return flow_path[path_index]
+    return ''
+
+
+def flow_path_strip(flow_path: List[str]) -> List[str]:
+    """ return copy of passed flow_path with all non-enter actions stripped from the end.
+
+    :param flow_path:           flow path list to strip.
+    :return:                    stripped flow path copy.
+    """
+    deep = len(flow_path)
+    while deep and flow_action(flow_path_id(flow_path, path_index=deep - 1)) != 'enter':
+        deep -= 1
+    return flow_path[:deep]
+
+
 def flow_popup_class_name(flow_id: str) -> str:
     """ determine name of the Popup class for the given flow id.
 
@@ -528,11 +575,7 @@ def flow_popup_class_name(flow_id: str) -> str:
     :return:                name of the Popup class. Please note that the action `open` will not be added
                             to the returned class name.
     """
-    flow, _index = flow_key_split(flow_id)
-    action, obj = flow_action_split(flow)
-    if action == 'open':
-        action = ''
-    return f'{snake_to_camel(obj)}{action.capitalize()}Popup'
+    return flow_class_name(flow_id, 'Popup')
 
 
 def id_of_flow(action: str, obj: str = "", key: str = "") -> str:
@@ -598,14 +641,14 @@ def replace_flow_action(flow_id: str, new_action: str):
 
 
 def update_tap_kwargs(widget: Any, popup_kwargs: Optional[EventKwargsType] = None, **tap_kwargs) -> EventKwargsType:
-    """ update widget's tap_kwargs property and return the updated dictionary (for kv rule of tap_kwargs).
+    """ update or simulate widget's tap_kwargs property and return the updated dictionary (for kv rule of tap_kwargs).
 
     :param widget:              widget with tap_kwargs property to be updated.
     :param popup_kwargs:        dict with items for to update popup_kwargs key of tap_kwargs
     :param tap_kwargs:          additional tap_kwargs items to update.
     :return:
     """
-    handle = dict(tap_kwargs=widget.tap_kwargs)
+    handle = dict(tap_kwargs=widget.tap_kwargs) if hasattr(widget, 'tap_kwargs') else dict()
     ensure_tap_kwargs_refs(handle, widget)
     new_kwargs = handle['tap_kwargs']
     if popup_kwargs:
@@ -710,25 +753,26 @@ class MainAppBase(ConsoleApp, ABC):
 
         return backup_root
 
-    def change_app_state(self, state_name: str, state_value: Any, send_event: bool = True, old_name: str = ''):
-        """ change app state to :paramref:`~change_app_state.state_value` in self.<state_name> and app_state dict item.
+    def change_app_state(self, app_state_name: str, state_value: Any, send_event: bool = True, old_name: str = ''):
+        """ change app state to :paramref:`~change_app_state.state_value` in self.<app_state_name> and app_states dict.
 
-        :param state_name:      name of the app state to change.
+        :param app_state_name:  name of the app state to change.
         :param state_value:     new value of the app state to change.
-        :param send_event:      pass False to prevent to send/call the on_<state_name> event to the main app instance.
-        :param old_name:        pass to add state to the main config file: old state name to rename/migrate or `UNSET`
-                                to only add a new app state variable with the name specified in :paramref:`.state_name`.
+        :param send_event:      pass False to prevent send/call of the main_app.on_<app_state_name> event.
+        :param old_name:        pass to add state to the main config file: old state name to rename/migrate or
+                                :data:`~ae.base.UNSET` to only add a new app state variable with the name specified in
+                                :paramref:`.app_state_name`.
         """
-        self.vpo(f"MainAppBase.change_app_state({state_name}, {state_value}, {send_event}, {old_name})"
+        self.vpo(f"MainAppBase.change_app_state({app_state_name}, {state_value!r}, {send_event}, {old_name!r})"
                  f" flow={self.flow_id} last={self._last_focus_flow_id} path={self.flow_path}")
 
-        self.change_observable(state_name, state_value, is_app_state=True)
+        self.change_observable(app_state_name, state_value, is_app_state=True)
 
         if old_name or old_name == UNSET:
-            self.set_var(state_name, state_value, section=APP_STATE_SECTION_NAME, old_name=old_name or "")
+            self.set_var(app_state_name, state_value, section=APP_STATE_SECTION_NAME, old_name=old_name or "")
 
         if send_event:
-            self.call_method('on_' + state_name)
+            self.call_method('on_' + app_state_name)
 
     def change_observable(self, name: str, value: Any, is_app_state: bool = False):
         """ change observable attribute/member/property in framework_app instance (and shadow copy in main app).
@@ -792,7 +836,7 @@ class MainAppBase(ConsoleApp, ABC):
                                 indexed widget the input focus (these exceptions are configurable via
                                 the list :data:`ACTIONS_CHANGING_FLOW_WITHOUT_CONFIRMATION`).
         """
-        self.vpo(f"MainAppBase.change_flow '{new_flow_id}'/'{self.flow_id}' path={self.flow_path} kwa={event_kwargs}")
+        self.vpo(f"MainAppBase.change_flow({new_flow_id!r}, {event_kwargs}) id={self.flow_id!r} path={self.flow_path}")
 
         action = flow_action(new_flow_id)
         if not self.call_method(flow_change_confirmation_event_name(new_flow_id), flow_key(new_flow_id), event_kwargs) \
@@ -872,6 +916,11 @@ class MainAppBase(ConsoleApp, ABC):
         """ dpi scaling factor - overwrite if the used GUI framework supports dpi scaling. """
         return 1.0
 
+    def close_popups(self):
+        """ close all opened popups (starting with the topmost popup). """
+        for popup in self.popups_opened():
+            popup.close()
+
     def find_image(self, image_name: str, height: float = 32.0, light_theme: bool = True) -> Optional[RegisteredFile]:
         """ find best fitting image in img app folder (see also :meth:`~MainAppBase.img_file` for easier usage).
 
@@ -910,29 +959,52 @@ class MainAppBase(ConsoleApp, ABC):
             return self.sound_files(sound_name)
         return None
 
+    def find_widget(self, match: Callable[[Any], bool]) -> Optional[Any]:
+        """ search the widget tree returning the first matching widget in reversed z-order (top-most first).
+
+        :param match:           callable called with the widget as argument, returning True if widget matches.
+        :return:                first found widget in reversed z-order (top-most widget first).
+        """
+        def child_wid(children):
+            """ bottom up search within children for a widget with matching attribute name and value. """
+            for widget in reversed(children):
+                found = child_wid(self.widget_children(widget))
+                if found:
+                    return found
+                if match(widget):
+                    return widget
+            return None
+
+        return child_wid(self.widget_children(self.framework_win))
+
     def flow_path_action(self, flow_path: Optional[List[str]] = None, path_index: int = -1) -> str:
         """ determine the action of the last/newest entry in the flow_path.
 
-        :param flow_path:       flow path to get the flow action from (default=self.flow_path).
-        :param path_index:      index in the flow_path.
-        :return:                flow action string or empty string if index does not exist.
+        :param flow_path:       optional flow path to get the flow action from (default=self.flow_path).
+        :param path_index:      optional index in the flow_path (default=-1).
+        :return:                flow action string or empty string if flow path is empty or index does not exist.
         """
         if flow_path is None:
             flow_path = self.flow_path
-        if len(flow_path) >= (abs(path_index) if path_index < 0 else path_index + 1):
-            return flow_action(flow_path[path_index])
-        return ''
+        return flow_action(flow_path_id(flow_path=flow_path, path_index=path_index))
 
-    def flow_path_strip(self, flow_path: List[str]) -> List[str]:
-        """ return flow_path copy with all non-enter actions stripped from the end.
+    def global_variables(self, **patches) -> Dict[str, Any]:
+        """ determine generic/most-needed global variables to evaluate expressions/macros.
 
-        :param flow_path:       flow path list to strip.
-        :return:                stripped flow path copy.
+        :param patches:         dict of variable names and values to add/replace on top of generic globals.
+        :return:                dict of global variables patched with :paramref:`.patches`.
         """
-        deep = len(flow_path)
-        while deep and self.flow_path_action(flow_path=flow_path, path_index=deep - 1) != 'enter':
-            deep -= 1
-        return flow_path[:deep]
+        glo_vars = {k: v for k, v in module_globals.items() if k not in HIDDEN_GLOBALS}
+        glo_vars.update((k, v) for k, v in globals().items() if k not in HIDDEN_GLOBALS)
+        glo_vars['app'] = self.framework_app
+        glo_vars['main_app'] = self
+        glo_vars['_add_base_globals'] = True        # instruct ae.inspector.try_eval to add ae.core globals
+
+        self.vpo(f"MainAppBase.global_variables patching {patches} over {glo_vars}")
+
+        glo_vars.update(**patches)
+
+        return glo_vars
 
     def img_file(self, image_name: str, font_size: Optional[float] = None, light_theme: Optional[bool] = None) -> str:
         """ shortcutting :meth:`~MainAppBase.find_image` method w/o bound property for to get image file path.
@@ -1026,6 +1098,11 @@ class MainAppBase(ConsoleApp, ABC):
             default_language(lang_code)
             self.set_var('lang_code', lang_code, section=APP_STATE_SECTION_NAME)  # add optional app state var to config
 
+    def mix_background_ink(self):
+        """ remix background ink if one of the basic back colours change. """
+        self.framework_app.mixed_back_ink = (sum(_) / len(_) for _ in zip(
+            self.flow_id_ink, self.flow_path_ink, self.selected_item_ink, self.unselected_item_ink))
+
     def on_app_build(self):
         """ default/fallback flow change confirmation event handler. """
         self.vpo("MainAppBase.on_app_build default/fallback event handler called - loading app resources")
@@ -1091,10 +1168,18 @@ class MainAppBase(ConsoleApp, ABC):
             popup_class = self.class_by_name(class_name)
             if popup_class:
                 popup_kwargs = event_kwargs.get('popup_kwargs', dict())
-                self.show_popup(popup_class, **popup_kwargs)
+                self.open_popup(popup_class, **popup_kwargs)
                 return True
         return False
 
+    def on_flow_id_ink(self):
+        """ redirect flow id back ink app state color change event handler to actualize mixed_back_ink. """
+        self.mix_background_ink()
+
+    def on_flow_path_ink(self):
+        """ redirect flow path back ink app state color change event handler to actualize mixed_back_ink. """
+        self.mix_background_ink()
+
     @staticmethod
     def on_flow_popup_close(_flow_key: str, _popup_args: EventKwargsType) -> bool:
         """ default popup close handler of FlowPopup widget, ensuring update of :attr:`flow_path`.
@@ -1128,6 +1213,14 @@ class MainAppBase(ConsoleApp, ABC):
         self.change_app_state('light_theme', light_theme)
         return True
 
+    def on_selected_item_ink(self):
+        """ redirect selected item back ink app state color change event handler to actualize mixed_back_ink. """
+        self.mix_background_ink()
+
+    def on_unselected_item_ink(self):
+        """ redirect unselected item back ink app state color change event handler to actualize mixed_back_ink. """
+        self.mix_background_ink()
+
     def play_beep(self):
         """ make a short beep sound, should be overwritten by GUI framework. """
         self.po(chr(7), "MainAppBase.BEEP")
@@ -1146,17 +1239,52 @@ class MainAppBase(ConsoleApp, ABC):
         """
         self.po(f"MainAppBase.play_vibrate {pattern}")
 
+    def open_popup(self, popup_class: Type, **popup_kwargs) -> Any:
+        """ open Popup or DropDown by calling `show` method of passed instance.
+
+        Overwrite this method if framework is using different method to open popup window or if
+        a widget in the Popup/DropDown need to get the input focus.
+
+        :param popup_class:     class of the Popup/DropDown widget/window.
+        :param popup_kwargs:    args for to instantiate and show/open the popup.
+        :return:                created and displayed/opened popup class instance.
+        """
+        self.dpo(f"MainAppBase.open_popup {popup_class} {popup_kwargs}")
+        popup_instance = popup_class(**popup_kwargs)
+        popup_instance.open()
+        return popup_instance
+
+    def popups_opened(self, classes: Tuple = ()) -> List:
+        """ determine all popup-like container widgets that are currently opened.
+
+        :param classes:         optional class filter - if not passed then only the widgets underneath win/root with an
+                                `open` method will be added. Pass tuple for to restrict found popup widgets to certain
+                                classes. like e.g. by passing (Popup, DropDown, FlowPopup) to get all popups of an
+                                ae/Kivy app.
+        :return:                list of opened/visible popup class instances that are children of either the
+                                root layout or the app window, ordered by their z-coordinate (most front widget last).
+        """
+        filter_func = (lambda _wg: isinstance(_wg, classes)) if classes else \
+            (lambda _wg: callable(getattr(_wg, 'open', None)))
+
+        popups = list()
+        for wid in self.framework_win.children + self.framework_root.children:
+            if filter_func(wid):
+                popups.append(wid)
+
+        return popups
+
     def retrieve_app_states(self) -> AppStateType:
         """ determine the state of a running app from the main app instance and return it as dict.
 
         :return:                dict with all app states available in the config files.
         """
-        app_state = dict()
+        app_states = dict()
         for key in self.app_state_keys():
-            app_state[key] = getattr(self, key)
+            app_states[key] = getattr(self, key)
 
-        self.dpo(f"MainAppBase.retrieve_app_states {app_state}")
-        return app_state
+        self.dpo(f"MainAppBase.retrieve_app_states {app_states}")
+        return app_states
 
     def run_app(self):
         """ startup main and framework applications. """
@@ -1182,9 +1310,8 @@ class MainAppBase(ConsoleApp, ABC):
         """
         err_msg = ""
 
-        app_state = self.retrieve_app_states()
-
-        for key, state in app_state.items():
+        app_states = self.retrieve_app_states()
+        for key, state in app_states.items():
             if isinstance(state, (list, dict)):
                 state = deepcopy(state)
 
@@ -1195,16 +1322,18 @@ class MainAppBase(ConsoleApp, ABC):
             if key == 'flow_id' and flow_action(state) != 'focus':
                 state = id_of_flow('')
             elif key == 'flow_path':
-                state = [flow_id for flow_id in state if flow_action(flow_id) == 'enter']
+                state = flow_path_strip(state)
 
             err_msg = self.set_var(key, state, section=APP_STATE_SECTION_NAME)
             self.vpo(f"MainAppBase.save_app_state {key}={state} {err_msg or 'OK'}")
             if err_msg:
                 break
+
         self.load_cfg_files()
 
         if self.debug_level:
             self.play_sound('error' if err_msg else 'debug_save')
+
         return err_msg
 
     def setup_app_states(self, app_states: AppStateType):
@@ -1232,21 +1361,6 @@ class MainAppBase(ConsoleApp, ABC):
                 self.call_method('on_app_state_version_upgrade', from_version)
             self.change_app_state(APP_STATE_VERSION_VAR_NAME, upgrade_version, send_event=False, old_name=UNSET)
 
-    def show_popup(self, popup_class: Type, **popup_kwargs) -> Any:
-        """ open Popup or DropDown by calling `show` method of passed instance.
-
-        Overwrite this method if framework is using different method to open popup window or if
-        a widget in the Popup/DropDown need to get the input focus.
-
-        :param popup_class:     class of the Popup/DropDown widget/window.
-        :param popup_kwargs:    args for to instantiate and show/open the popup.
-        :return:                created and displayed/opened popup class instance.
-        """
-        self.dpo(f"MainAppBase.show_popup {popup_class} {popup_kwargs}")
-        popup_instance = popup_class(**popup_kwargs)
-        popup_instance.show()
-        return popup_instance
-
     def stop_app(self, exit_code: int = 0):
         """ quit this application.
 
@@ -1277,39 +1391,50 @@ class MainAppBase(ConsoleApp, ABC):
         return cfg_parser.getint(APP_STATE_SECTION_NAME, APP_STATE_VERSION_VAR_NAME, fallback=0)
 
     def widget_by_attribute(self, att_name: str, att_value: str) -> Optional[Any]:
-        """ determine the widget referenced by the passed attribute name and value.
+        """ determine the first (top-most) widget having the passed attribute name and value.
 
         :param att_name:        name of the attribute of the searched widget.
         :param att_value:       attribute value of the searched widget.
-        :return:                widget that has the specified attribute with the specified value.
+        :return:                widget that has the specified attribute with the specified value or None if not found.
         """
-        def child_wid(children):
-            """ bottom up search within children for a widget with matching attribute name and value. """
-            for widget in children:
-                found = child_wid(widget.children)
-                if found:
-                    return found
-                if getattr(widget, att_name, None) == att_value:
-                    return widget
-            return None
-
-        return child_wid(self.framework_win.children)
+        return self.find_widget(lambda widget: getattr(widget, att_name, None) == att_value)
 
     def widget_by_flow_id(self, flow_id: str) -> Optional[Any]:
-        """ determine the widget having the passed flow_id.
+        """ determine the first (top-most) widget having the passed flow_id.
 
         :param flow_id:         flow id value of the searched widget's `tap_flow_id` attribute.
-        :return:                widget that has a `tap_flow_id` attribute with the value of the passed flow id.
+        :return:                widget that has a `tap_flow_id` attribute with the value of the passed flow id
+                                or None if not found.
         """
         return self.widget_by_attribute('tap_flow_id', flow_id)
 
-    def widget_by_state_name(self, state_name: str) -> Optional[Any]:
-        """ determine the widget having the passed app state name (app_state_name).
+    def widget_by_app_state_name(self, app_state_name: str) -> Optional[Any]:
+        """ determine the first (top-most) widget having the passed app state name (app_state_name).
 
-        :param state_name:      app state name of the widget's `app_state_name` attribute.
-        :return:                widget that has a `app_state_name` attribute with the passed app state name.
+        :param app_state_name:  app state name of the widget's `app_state_name` attribute.
+        :return:                widget that has a `app_state_name` attribute with the passed app state name
+                                or None if not found.
         """
-        return self.widget_by_attribute('app_state_name', state_name)
+        return self.widget_by_attribute('app_state_name', app_state_name)
+
+    def widget_children(self, wid: Any, only_visible: bool = False) -> List:
+        """ determine the children of widget or its container (if exists) in z-order (top-most last).
+
+        :param wid:             widget to determine the children from.
+        :param only_visible:    pass True to only return visible widgets.
+        :return:                list of children widgets of the passed widget.
+        """
+        wid_visible = self.widget_visible
+        return [chi for chi in getattr(wid, 'container', wid).children if not only_visible or wid_visible(chi)]
+
+    @staticmethod
+    def widget_visible(wid: Any) -> bool:
+        """ determine if the passed widget is visible (has width and height and (visibility or opacity) set).
+
+        :param wid:             widget to determine visibility of.
+        :return:                True if widget is visible (or visibility cannot be determined), False if hidden.
+        """
+        return bool(wid.width and wid.height and getattr(wid, 'visible', True) and getattr(wid, 'opacity', True))
 
     def win_pos_size_change(self, *win_pos_size):
         """ screen resize handler (called on window resize or when app will exit/stop via closed event.
@@ -1325,5 +1450,7 @@ class MainAppBase(ConsoleApp, ABC):
 register_package_images()   # register base image files of this portion
 register_package_sounds()   # register base sound files of this portion
 
+# bind imported but unused names, to be available in :meth:`MainAppBase.global_variables`
+_d_ = (get_f_string, get_text, os_platform, path_name, placeholder_path)
 module_globals = globals()
 #: used. e.g. by :mod:`ae.gui_help` for execution/evaluation of dynamic code, expressions and f-strings
